from __future__ import annotations

import functools
import signal
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Generic

import anyio
from anyio import CancelScope, create_memory_object_stream, from_thread
from pydantic import ValidationError

from .broker import Broker, BulkMessage, R
from .consumer import ChannelConsumer, Consumer, ConsumerGroup
from .decoder import DEFAULT_DECODER
from .encoder import DEFAULT_ENCODER
from .exceptions import DecodeError, Fail, Retry, Skip
from .logging import LoggerMixin
from .models import CloudEvent, Publishes
from .types import (
    Decoder,
    Encoder,
    Lifespan,
    Message,
    MiddlewareType,
    P,
    PreparedMessage,
    Publisher,
)
from .utils import to_float

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Sequence

    from anyio.abc import TaskGroup
    from anyio.streams.memory import MemoryObjectReceiveStream

    from .middleware import Middleware


@asynccontextmanager
async def nullcontext(_: Service) -> AsyncIterator[None]:
    yield


class Service(Generic[Message, R], LoggerMixin):
    """Logical group of consumers. Provides group (queue) name and handles versioning."""

    default_middlewares: ClassVar[list[MiddlewareType]] = []

    def __init__(
        self,
        name: str,
        broker: Broker[Message, R],
        encoder: Encoder = DEFAULT_ENCODER,
        decoder: Decoder = DEFAULT_DECODER,
        title: str | None = None,
        version: str = "0.1.0",
        description: str = "",
        lifespan: Lifespan = nullcontext,
        tags_metadata: list[dict[str, Any]] | None = None,
        publishes: list[Publishes] | None = None,
        async_api_extra: dict[str, Any] | None = None,
        state: dict[type | str, Any] | None = None,
        **options: Any,
    ) -> None:
        self.broker = broker
        self.name = name
        self.encoder = encoder
        self.decoder = decoder
        self.title = title or name.title()
        self.version = version
        self.description = description
        self.tags_metadata = tags_metadata or []
        self.consumer_group = ConsumerGroup()
        self.subscribe = self.consumer_group.subscribe
        self.middlewares: list[Middleware] = []
        for m in self.default_middlewares:
            self.add_middleware(m)
        self.lifespan = lifespan
        self.publishes = publishes or []
        self.async_api_extra = async_api_extra or {}
        self.state = state or {}
        self.state[Publisher] = self.publish
        self.options = options
        self.default_action = getattr(self, self.broker.default_on_exc)

    def add_middleware(
        self, middleware: MiddlewareType[P], *args: P.args, **kwargs: P.kwargs
    ) -> None:
        self.middlewares.append(middleware(self, *args, **kwargs))

    @property
    def add_consumer_group(self) -> Callable[[ConsumerGroup], None]:
        return self.consumer_group.add_consumer_group

    @property
    def consumers(self) -> dict[str, Consumer]:
        return self.consumer_group.consumers

    async def send(
        self,
        data: Any,
        type: type[CloudEvent] | str = CloudEvent,
        headers: dict[str, str] | None = None,
        encoder: Encoder | None = None,
        **kwargs: Any,
    ) -> R:
        if isinstance(type, str):
            kwargs["type"] = type
            type = CloudEvent
        ce = type.new(data, type=type, source=self.name, headers=headers, **kwargs)
        if headers:
            ce.headers.update(headers)
        return await self.publish(ce, encoder=encoder)

    def prepare_message(
        self,
        message: CloudEvent,
        topic: str | None = None,
        encoder: Encoder | None = None,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> PreparedMessage:
        message_topic = topic or message.topic
        encoder = encoder or self.encoder
        if message.source is None:
            message.source = self.name
        message.content_type = encoder.CONTENT_TYPE
        message.headers["Content-Type"] = encoder.CONTENT_TYPE
        if topic:
            message.headers["Destination"] = topic
        if headers:
            message.headers.update(headers)
        body = encoder.encode(message)
        message_kwargs = {
            f"message_{k}": v
            for k, v in message.model_dump(
                exclude_none=True, by_alias=False, exclude={"topic", "data"}
            ).items()
        }
        message_kwargs.update(kwargs)
        return message_topic, body, message_kwargs

    def publish_sync(
        self,
        message: CloudEvent,
        topic: str | None = None,
        headers: dict[str, str] | None = None,
        encoder: Encoder | None = None,
        **kwargs: Any,
    ) -> R:
        fn = functools.partial(
            self.publish,
            message,
            topic=topic,
            headers=headers,
            encoder=encoder,
            **kwargs,
        )
        return from_thread.run(fn)

    async def publish(
        self,
        message: CloudEvent,
        topic: str | None = None,
        headers: dict[str, str] | None = None,
        encoder: Encoder | None = None,
        **kwargs: Any,
    ) -> R:
        message_topic, body, message_kwargs = self.prepare_message(
            message, topic, encoder, headers=headers, **kwargs
        )
        topic = topic or message.topic
        await self.dispatch_before("publish", message=message, **kwargs)
        res = await self.broker.publish(
            message_topic, body, headers=message.headers, **message_kwargs
        )
        await self.dispatch_after("publish", message=message, **kwargs)
        return res

    def bulk_publish_sync(
        self,
        messages: Sequence[CloudEvent],
        *,
        topic: str | None = None,
        headers: dict[str, str] | None = None,
        encoder: Encoder | None = None,
        **kwargs: Any,
    ) -> None:
        fn = functools.partial(
            self.bulk_publish,
            topic=topic,
            headers=headers,
            encoder=encoder,
            **kwargs,
        )
        from_thread.run(fn)

    async def bulk_publish(
        self,
        messages: Sequence[CloudEvent],
        *,
        topic: str | None = None,
        headers: dict[str, str] | None = None,
        encoder: Encoder | None = None,
        **kwargs: Any,
    ) -> None:
        bulk_messages: list[BulkMessage] = []
        for message in messages:
            message_topic, body, message_kwargs = self.prepare_message(
                message, topic, encoder, headers=headers, **kwargs
            )
            await self.dispatch_before("publish", message=message, **kwargs)
            msg = BulkMessage(message_topic, body, message.headers, message_kwargs)
            bulk_messages.append(msg)

        await self.broker.bulk_publish(bulk_messages, topic=topic)

        for message in messages:
            await self.dispatch_after("publish", message=message, **kwargs)

    async def connect(self) -> None:
        await self.dispatch_before("broker_connect")
        await self.broker.connect()
        await self.dispatch_after("broker_connect")

    async def disconnect(self) -> None:
        await self.dispatch_before("broker_disconnect")
        await self.broker.disconnect()
        await self.dispatch_after("broker_disconnect")

    async def start_consumers(self, tg: TaskGroup) -> None:
        for consumer in self.consumers.values():
            consumer.maybe_set_publisher(self.publish)
            await self.dispatch_before("consumer_start", consumer=consumer)
            send_stream, receive_stream = create_memory_object_stream[Any](
                consumer.concurrency,
            )

            tg.start_soon(self.broker.sender, self.name, consumer, send_stream)

            for i in range(consumer.concurrency):
                self.logger.info("Starting consumer %s task %s", consumer.name, i)
                tg.start_soon(
                    self.receiver,
                    consumer,
                    receive_stream.clone(),
                    name=f"{consumer.name}:{i+1}",
                )
            await self.dispatch_after("consumer_start", consumer=consumer)

    async def run(self, enable_signal_handler: bool = True) -> None:
        async with self.lifespan(self) as state:
            if state:
                self.state.update(state)

            await self.connect()
            try:
                async with anyio.create_task_group() as tg:
                    if enable_signal_handler:
                        tg.start_soon(self.watch_for_signals, tg.cancel_scope)
                    await self.start_consumers(tg)
            finally:
                with anyio.move_on_after(5, shield=True):
                    await self.disconnect()

    @asynccontextmanager
    async def context(self, enable_signal_handler: bool = False) -> AsyncIterator[None]:
        async with self.lifespan(self) as state:
            if state:
                self.state.update(state)
            try:
                await self.connect()
                async with anyio.create_task_group() as tg:
                    if enable_signal_handler:
                        tg.start_soon(self.watch_for_signals, tg.cancel_scope)
                    await self.start_consumers(tg)
                    yield
            finally:
                with anyio.move_on_after(5, shield=True):
                    await self.disconnect()

    async def watch_for_signals(self, scope: CancelScope) -> None:
        with anyio.open_signal_receiver(signal.SIGINT, signal.SIGTERM) as signals:
            async for signum in signals:
                self.logger.info("Received signal %s", signum.name)
                scope.cancel()

    async def _dispatch(self, event: str, **kwargs: Any) -> None:
        message = kwargs.get("message")
        for middleware in self.middlewares:
            if message and (
                middleware.requires is not None
                and not isinstance(message, middleware.requires)
            ):
                self.logger.debug(
                    "Skipping event %s for middleware %s",
                    event,
                    type(middleware).__name__,
                )
                continue
            method = getattr(middleware, event, None)
            if method is None:
                self.logger.debug(
                    "Method %s not found in middleware %s",
                    event,
                    type(middleware).__name__,
                )
                continue

            await method(**kwargs)

    async def dispatch_before(self, event: str, **kwargs: Any) -> None:
        await self._dispatch(f"before_{event}", **kwargs)

    async def dispatch_after(self, event: str, **kwargs: Any) -> None:
        await self._dispatch(f"after_{event}", **kwargs)

    async def receiver(
        self,
        consumer: Consumer,
        receive_stream: MemoryObjectReceiveStream[Message],
    ) -> None:
        consumer_timeout = to_float(
            consumer.timeout or self.broker.default_consumer_timeout,
        )
        decoder = consumer.decoder or self.decoder

        async with receive_stream:
            async for raw_message in receive_stream:
                await self._process(
                    consumer,
                    raw_message,
                    decoder,
                    consumer_timeout,
                )

    async def ack(self, consumer: Consumer, message: Message) -> None:
        await self.dispatch_before("ack", consumer=consumer, raw_message=message)
        await self.broker.ack(message)
        await self.dispatch_after("ack", consumer=consumer, raw_message=message)

    async def nack(
        self,
        consumer: Consumer,
        message: Message,
        delay: int | None = None,
    ) -> None:
        await self.dispatch_after(
            "nack",
            consumer=consumer,
            raw_message=message,
        )
        await self.broker.nack(message, delay=delay)
        await self.dispatch_after(
            "nack",
            consumer=consumer,
            raw_message=message,
        )

    async def _process(
        self,
        consumer: Consumer,
        raw_message: Message,
        decoder: Decoder,
        timeout: float,
    ) -> None:
        exc: Exception | None = None
        result = None

        try:
            data, headers = self.broker.decode_message(raw_message)
            message = decoder.decode(data, consumer.event_type)
            message.set_context(self, raw_message, headers)
        except (DecodeError, ValidationError) as e:
            self.logger.exception(
                "Failed to validate message %s.",
                raw_message,
                exc_info=e,
            )
            if self.broker.should_nack(raw_message):
                await self.nack(
                    consumer,
                    raw_message,
                    delay=self.broker.validate_error_delay,
                )
            else:
                await self.ack(consumer, raw_message)
            return

        try:
            await self.dispatch_before(
                "process_message",
                consumer=consumer,
                message=message,
            )
            self.logger.info(
                "Running consumer %s with message %s", consumer.name, message.id
            )
            with anyio.fail_after(timeout):
                result = await consumer.process(message)
        except Exception as e:
            exc = e
        except anyio.get_cancelled_exc_class():
            with anyio.move_on_after(1, shield=True):
                # directly call nack skipping all middlewares
                await self.broker.nack(raw_message)
            raise
        await self._handle_message_finalization(consumer, message, result, exc)

    async def _handle_message_finalization(
        self,
        consumer: Consumer,
        message: CloudEvent,
        result: Any,
        exc: Exception | None,
    ) -> None:
        try:
            await self.dispatch_after(
                "process_message",
                consumer=consumer,
                message=message,
                result=result,
                exc=exc,
            )
        except Exception as e:
            exc = e

        if exc is None:
            await self.ack(consumer, message.raw)
            return

        if isinstance(exc, Retry):
            await self.dispatch_after(
                "retry_message",
                consumer=consumer,
                message=message,
                exc=exc,
            )
            await self.nack(consumer, message.raw, delay=exc.delay)
            return

        if isinstance(exc, Skip):
            await self.dispatch_after(
                "skip_message",
                consumer=consumer,
                message=message,
                exc=exc,
            )
            await self.ack(consumer, message.raw)
            return

        if isinstance(exc, Fail):
            await self.dispatch_after(
                "fail_message",
                consumer=consumer,
                message=message,
                exc=exc,
            )
            await self.ack(consumer, message.raw)
            return

        await self.default_action(consumer, message.raw)

    @asynccontextmanager
    async def subscription(
        self,
        event_type: type[CloudEvent] = CloudEvent,
        topic: str | None = None,
        **options: Any,
    ) -> AsyncIterator[
        MemoryObjectReceiveStream[tuple[CloudEvent, Callable[[], None]]]
    ]:
        """Async with service.subscription(MyEvent, topic="example.topic") as subscription:
        async for event, ack in subscription:
            print(event)
            ack().
        """
        send_stream, receive_stream = create_memory_object_stream[Any](1)
        consumer_send, user_receive = create_memory_object_stream[
            tuple[CloudEvent, Callable[[], None]]
        ](1)
        options["dynamic"] = True
        consumer = ChannelConsumer(
            channel=consumer_send,
            event_type=event_type,
            topic=topic,
            **options,
        )

        async with anyio.create_task_group() as tg, consumer_send, user_receive:
            tg.start_soon(self.broker.sender, self.name, consumer, send_stream)
            tg.start_soon(self.receiver, consumer, receive_stream)
            yield user_receive



from __future__ import annotations

from logging import INFO
from time import perf_counter
from typing import TYPE_CHECKING, Any, Literal

from eventiq.middleware import CloudEventType, Middleware

if TYPE_CHECKING:
    from eventiq import Consumer, Service
    from eventiq.types import ID


class PerfCounterMiddleware(Middleware[CloudEventType]):
    def __init__(self, service: Service, log_level: int = INFO) -> None:
        super().__init__(service)
        self.log_level = log_level
        self._receive_registry: dict[tuple[str, ID], float] = {}
        self._publish_registry: dict[ID, float] = {}

    def _log_elapsed_time(
        self, operation: Literal["processed", "published"], start_time: float
    ) -> None:
        elapsed = perf_counter() - start_time
        self.logger.log(
            self.log_level, "Message %s in %.2f seconds.", operation, elapsed
        )

    async def before_process_message(
        self, *, consumer: Consumer, message: CloudEventType
    ) -> None:
        self._receive_registry[(consumer.name, message.id)] = perf_counter()

    async def after_process_message(
        self,
        *,
        consumer: Consumer,
        message: CloudEventType,
        result: Any = None,
        exc: Exception | None = None,
    ) -> None:
        start_time = self._receive_registry.pop(
            (consumer.name, message.id), perf_counter()
        )
        self._log_elapsed_time("processed", start_time)

    async def before_publish(self, *, message: CloudEventType, **kwargs: Any) -> None:
        self._publish_registry[message.id] = perf_counter()

    async def after_publish(self, *, message: CloudEventType, **kwargs: Any) -> None:
        start_time = self._publish_registry.pop(message.id, perf_counter())
        self._log_elapsed_time("published", start_time)



from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import Protocol

from eventiq import CloudEvent, Service
from eventiq.backends.nats import JetStreamBroker
from eventiq.types import State

broker = JetStreamBroker(url="nats://localhost:4222")


class UserRepository(Protocol):
    async def get_usernames(self) -> list[str]: ...


class FakeUserRepository:
    def __init__(self, data) -> None:
        self.data = data

    async def get_usernames(self) -> list[str]:
        return self.data


@asynccontextmanager
async def lifespan(_service) -> AsyncIterator[State]:
    print("setup dependencies, connect to db, etc")
    yield {UserRepository: FakeUserRepository(["user1", "user2"])}
    print("cleanup")


service = Service(name="example-service", broker=broker, lifespan=lifespan)


@service.subscribe(topic="test.topic", concurrency=10)
async def example_run(message: CloudEvent, *, repository: UserRepository):
    print(f"Received Message {message.id} with data: {message.data}")
    assert isinstance(repository, FakeUserRepository)
    usernames = await repository.get_usernames()
    assert usernames == ["user1", "user2"]



import asyncio

from eventiq import CloudEvent, Middleware, Service
from eventiq.backends.nats import JetStreamBroker


class SendMessageMiddleware(Middleware):
    async def after_broker_connect(self):
        print(f"After service start, running with {self.service.broker}")
        for i in range(1, 10):
            message = CloudEvent.new({"counter": i}, topic="test.topic")
            await self.service.publish(message)
        print("Published messages(s)")


broker = JetStreamBroker(url="nats://localhost:4222")

service = Service(
    name="example-service",
    broker=broker,
)
service.add_middleware(SendMessageMiddleware)


@service.subscribe(topic="test.topic", concurrency=2)
async def example_run(message: CloudEvent):
    print("Received Message", message.id, "with data:", message.data)
    await asyncio.sleep(5)



from eventiq.decoder import DEFAULT_DECODER
from eventiq.encoder import DEFAULT_ENCODER


def test_default_encoder_decoder(ce):
    data = DEFAULT_ENCODER.encode(ce)
    parsed = DEFAULT_DECODER.decode(data, as_type=type(ce))
    assert parsed == ce



from __future__ import annotations

from datetime import datetime, timedelta
from itertools import chain
from typing import TYPE_CHECKING, Annotated, Any
from weakref import WeakValueDictionary

from aiokafka import AIOKafkaConsumer, AIOKafkaProducer, ConsumerRecord, TopicPartition
from anyio import move_on_after
from pydantic import AnyUrl, Field, UrlConstraints

from eventiq.broker import UrlBroker
from eventiq.settings import UrlBrokerSettings
from eventiq.utils import utc_now

if TYPE_CHECKING:
    from anyio.streams.memory import MemoryObjectSendStream

    from eventiq import Consumer
    from eventiq.types import ID, DecodedMessage

KafkaUrl = Annotated[AnyUrl, UrlConstraints(allowed_schemes=["kafka"])]


class KafkaSettings(UrlBrokerSettings[KafkaUrl]):
    consumer_options: dict[str, Any] = Field({})


class KafkaBroker(UrlBroker[ConsumerRecord, None]):
    """
    Kafka backend
    :param consumer_options: extra options (defaults) for AIOKafkaConsumer
    :param kwargs: Broker base class parameters
    """

    WILDCARD_MANY = "*"
    WILDCARD_ONE = r"\w+"

    Settings = KafkaSettings
    protocol = "kafka"

    def __init__(
        self,
        *,
        consumer_options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.consumer_options = consumer_options or {}
        self._publisher = None
        self._subcsribers: WeakValueDictionary[int, AIOKafkaConsumer] = (
            WeakValueDictionary()
        )

    @staticmethod
    def decode_message(raw_message: ConsumerRecord) -> DecodedMessage:
        data = raw_message.value or b""
        headers = {k: str(v) for k, v in raw_message.headers}
        return data, headers

    @staticmethod
    def get_message_metadata(
        raw_message: ConsumerRecord,
    ) -> dict[str, str]:
        metadata = {
            "messaging.kafka.message.offset": str(raw_message.offset),
            "messaging.kafka.destination.partition": str(raw_message.partition),
        }
        if raw_message.key:
            metadata["messaging.kafka.message.key"] = str(raw_message.key)
        return metadata

    @property
    def is_connected(self) -> bool:
        return True

    def should_nack(self, raw_message: ConsumerRecord) -> bool:
        return (
            raw_message.timestamp
            < (utc_now() + timedelta(seconds=self.validate_error_delay)).timestamp()
        )

    async def sender(
        self,
        group: str,
        consumer: Consumer,
        send_stream: MemoryObjectSendStream[ConsumerRecord],
    ) -> None:
        subscriber = AIOKafkaConsumer(
            group_id=f"{group}:{consumer.name}",
            bootstrap_servers=self.url,
            enable_auto_commit=False,
            **consumer.options.get("kafka_consumer_options", self.consumer_options),
        )
        subscriber.subscribe(pattern=self.format_topic(consumer.topic))
        await subscriber.start()
        timeout_ms = consumer.options.get("timeout_ms", 600)

        try:
            async with send_stream:
                while True:
                    result = await subscriber.getmany(timeout_ms=timeout_ms)
                    for message in chain.from_iterable(result.values()):
                        self._subcsribers[id(message)] = subscriber
                        await send_stream.send(message)
        finally:
            with move_on_after(1, shield=True):
                await subscriber.stop()

                if consumer.dynamic:
                    subscriber.unsubscribe()

    async def ack(self, raw_message: ConsumerRecord) -> None:
        subscriber = self._subcsribers.pop(id(raw_message), None)
        if subscriber:
            await subscriber.commit(
                {
                    TopicPartition(
                        raw_message.topic, raw_message.partition
                    ): raw_message.offset + 1
                }
            )

    async def nack(self, raw_message: ConsumerRecord, delay: int | None = None) -> None:
        self._subcsribers.pop(id(raw_message), None)

    async def disconnect(self) -> None:
        if self._publisher:
            await self._publisher.stop()

    @property
    def publisher(self) -> AIOKafkaProducer:
        if self._publisher is None:
            raise self.connection_error
        return self._publisher

    async def connect(self) -> None:
        if self._publisher is None:
            _publisher = AIOKafkaProducer(
                bootstrap_servers=self.url, **self.connection_options
            )
            self._publisher = _publisher
            await _publisher.start()

    async def publish(
        self,
        topic: str,
        body: bytes,
        *,
        headers: dict[str, str],
        message_id: ID,
        message_time: datetime,
        timestamp_ms: int | None = None,
        partition: int | None = None,
        **kwargs: Any,
    ) -> None:
        if timestamp_ms is None:
            timestamp_ms = int(message_time.timestamp() * 1000)
        await self.publisher.send(
            topic=topic,
            value=body,
            key=str(message_id),
            partition=partition,
            timestamp_ms=timestamp_ms,
            headers=headers,
        )



from __future__ import annotations


class EventiqError(Exception):
    """Base exception for Eventiq."""


class BrokerError(EventiqError):
    """Base Exception for broker related errors."""


class BrokerConnectionError(BrokerError):
    """Broker connection error."""


class EncodeError(EventiqError):
    """Error encoding message."""


class DecodeError(EventiqError):
    """Error decoding message."""


class MessageError(EventiqError):
    """Base message processing error."""

    def __init__(self, reason: str) -> None:
        self.reason = reason

    def __str__(self) -> str:
        return f"{self.__class__.__name__}: {self.reason}"


class Skip(MessageError):
    """Raise exception to skip message without processing and/or retrying."""


class Fail(MessageError):
    """Fail message without retrying."""


class Retry(MessageError):
    """Utility exception for retrying message.
    RetryMiddleware must be added.
    """

    def __init__(self, reason: str | None = None, delay: int | None = None) -> None:
        super().__init__(reason or "unknown")
        self.delay = delay



import os

import pytest

from eventiq.backends.kafka import KafkaBroker
from eventiq.backends.nats import JetStreamBroker, NatsBroker
from eventiq.backends.rabbitmq import RabbitmqBroker
from eventiq.backends.redis import RedisBroker
from eventiq.broker import Broker

backends = (NatsBroker, JetStreamBroker, KafkaBroker, RabbitmqBroker, RedisBroker)


@pytest.mark.parametrize("broker", backends)
def test_is_subclass(broker):
    assert issubclass(broker, Broker)


@pytest.mark.parametrize("broker", backends)
def test_from_env(broker):
    os.environ["BROKER_URL"] = f"{broker.protocol}://localhost:1111"
    broker.from_env()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from pydantic import RootModel

from .broker import Broker, Message, R
from .middleware import Middleware

if TYPE_CHECKING:
    from eventiq import Consumer, Service

    from .models import CloudEvent


class AnyModel(RootModel[Any]):
    pass


class ResultBackend(Broker[Message, R], ABC):
    async def init_storage(self) -> None:
        pass

    @abstractmethod
    async def store_result(self, key: str, result: bytes) -> None:
        raise NotImplementedError

    @abstractmethod
    async def get_result(self, key: str) -> bytes | None:
        raise NotImplementedError


class ResultBackendMiddleware(Middleware):
    def __init__(self, service: Service) -> None:
        super().__init__(service)
        if not isinstance(service.broker, ResultBackend):
            err = f"Broker type must be ResultBackend. Got {type(service.broker).__name__}"
            raise TypeError(err)

        self.result_backend: ResultBackend = service.broker

    async def after_process_message(
        self,
        *,
        consumer: Consumer,
        message: CloudEvent,
        result: Any | None = None,
        exc: Exception | None = None,
    ) -> None:
        if not all([result, exc is None, consumer.store_results]):
            return
        encoder = consumer.encoder or self.service.encoder
        data = encoder.encode(AnyModel(result))
        await self.result_backend.store_result(
            f"{self.service.name}:{message.id}",
            data,
        )



from __future__ import annotations

import importlib
from types import ModuleType
from typing import TYPE_CHECKING, Annotated, Any, TypeVar

from pydantic_core import PydanticCustomError, core_schema

AnyType = TypeVar("AnyType")


def import_from_string(path: str) -> Any:
    module_name, _, obj = path.partition(":")
    module = importlib.import_module(module_name)
    try:
        return getattr(module, obj)
    except AttributeError:
        msg = f"{module_name} has no object {obj}"
        raise ImportError(msg) from None


def import_string_validator(value: Any) -> Any:
    if isinstance(value, str):
        try:
            return import_from_string(value)
        except ImportError as e:
            msg = "import_error"
            raise PydanticCustomError(
                msg,
                "Invalid python path: {error}",
                {"error": str(e)},
            ) from None
    else:
        # otherwise we just return the value and let the next validator do the rest of the work
        return value


if TYPE_CHECKING:
    from pydantic import GetCoreSchemaHandler

    ImportedType = Annotated[AnyType, ...]
else:

    class ImportedType:
        @classmethod
        def __class_getitem__(cls, item: AnyType) -> AnyType:
            return Annotated[item, cls()]

        @classmethod
        def __get_pydantic_core_schema__(
            cls,
            source: type[Any],
            handler: GetCoreSchemaHandler,
        ) -> core_schema.CoreSchema:
            serializer = core_schema.plain_serializer_function_ser_schema(
                cls._serialize,
                when_used="json",
            )
            if cls is source:
                # Treat bare usage of ImportString (`schema is None`) as the same as ImportString[Any]
                return core_schema.no_info_plain_validator_function(
                    function=import_string_validator,
                    serialization=serializer,
                )
            return core_schema.no_info_before_validator_function(
                function=import_string_validator,
                schema=handler(source),
                serialization=serializer,
            )

        @staticmethod
        def _serialize(v: Any) -> str:
            if isinstance(v, ModuleType):
                return v.__name__
            if hasattr(v, "__module__") and hasattr(v, "__name__"):
                return f"{v.__module__}:{v.__name__}"
            return v

        def __repr__(self) -> str:
            return "ImportedType"



from __future__ import annotations

import asyncio
import re
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any

from eventiq.broker import Broker
from eventiq.settings import BrokerSettings
from eventiq.utils import utc_now

if TYPE_CHECKING:
    from anyio.streams.memory import MemoryObjectSendStream

    from eventiq import Consumer
    from eventiq.types import DecodedMessage


@dataclass
class StubMessage:
    data: bytes
    queue: asyncio.Queue
    event: asyncio.Event
    headers: dict[str, str] = field(default_factory=dict)


class StubSettings(BrokerSettings):
    wait_on_publish: bool = True


class StubBroker(Broker[StubMessage, dict[str, asyncio.Event]]):
    """This is in-memory implementation of a broker class, mainly designed for testing."""

    Settings = StubSettings

    protocol = ""

    WILDCARD_ONE = r"\w+"
    WILDCARD_MANY = "*"

    def __init__(
        self,
        *,
        wait_on_publish: bool = True,
        queue_max_size: int = 100,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.queue_max_size = queue_max_size
        self.topics: dict[str, asyncio.Queue[StubMessage]] = defaultdict(
            self.queue_factory,
        )
        self.wait_on_publish = wait_on_publish
        self._delay_queue: asyncio.Queue[tuple[StubMessage, datetime]] = asyncio.Queue(
            1000
        )
        self._connected = False
        self._delay_task: asyncio.Task | None = None

    @property
    def is_connected(self) -> bool:
        return self._connected

    def queue_factory(self) -> asyncio.Queue[StubMessage]:
        return asyncio.Queue(maxsize=self.queue_max_size)

    def get_info(self) -> dict[str, str]:
        return {"host": "localhost"}

    @staticmethod
    def decode_message(raw_message: StubMessage) -> DecodedMessage:
        return raw_message.data, raw_message.headers

    async def sender(
        self,
        group: str,
        consumer: Consumer,
        send_stream: MemoryObjectSendStream[StubMessage],
    ) -> None:
        queue = self.topics[self.format_topic(consumer.topic)]
        async with send_stream:
            while self._connected:
                message = await queue.get()
                await send_stream.send(message)

    async def _delay_queue_worker(self) -> None:
        while self._connected:
            raw_message, not_before = await self._delay_queue.get()
            if not_before > utc_now():
                await self._delay_queue.put((raw_message, not_before))
            else:
                await raw_message.queue.put(raw_message)
            self._delay_queue.task_done()
            await asyncio.sleep(0.1)

    async def connect(self) -> None:
        self._connected = True
        self._delay_task = asyncio.create_task(self._delay_queue_worker())
        await asyncio.sleep(0)

    async def disconnect(self) -> None:
        self._connected = False
        if self._delay_task:
            self._delay_task.cancel()
        self.topics.clear()

    async def publish(
        self,
        topic: str,
        body: bytes,
        *,
        headers: dict[str, str],
        **kwargs: Any,
    ) -> dict[str, asyncio.Event]:
        response = {}
        for target_topic, queue in self.topics.items():
            if re.fullmatch(topic, target_topic):
                event = asyncio.Event()
                msg = StubMessage(data=body, queue=queue, event=event, headers=headers)
                await queue.put(msg)
                response[topic] = event
                if self.wait_on_publish:
                    await event.wait()
        return response

    async def ack(self, raw_message: StubMessage) -> None:
        raw_message.queue.task_done()
        raw_message.event.set()

    async def nack(self, raw_message: StubMessage, delay: int | None = None) -> None:
        if delay and delay > 0:
            not_before = utc_now() + timedelta(seconds=delay)
            await self._delay_queue.put((raw_message, not_before))
        else:
            await raw_message.queue.put(raw_message)
        raw_message.queue.task_done()
        raw_message.event.set()



from __future__ import annotations

import inspect
import socket
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    overload,
)
from uuid import uuid4

import anyio
from typing_extensions import Concatenate, Unpack

from .dependencies import resolved_func
from .logging import get_logger
from .types import CloudEventType, P, RetryStrategy
from .utils import is_async_callable, resolve_message_type_hint, to_async, to_float

if TYPE_CHECKING:
    from collections.abc import Awaitable

    from anyio.streams.memory import MemoryObjectSendStream

    from .models import Publishes
    from .types import (
        ConsumerGroupOptions,
        Decoder,
        Encoder,
        MessageHandler,
        Parameter,
        Publisher,
        Timeout,
    )


class Consumer(ABC, Generic[CloudEventType]):
    """Base consumer class."""

    def __init__(
        self,
        *,
        name: str,
        event_type: type[CloudEventType],
        topic: str | None = None,
        concurrency: int = 1,
        timeout: Timeout | None = None,
        description: str | None = None,
        encoder: Encoder | None = None,
        decoder: Decoder | None = None,
        retry_strategy: RetryStrategy | None = None,
        dynamic: bool = False,
        store_results: bool = False,
        tags: list[str] | None = None,
        publishes: list[Publishes] | None = None,
        parameters: dict[str, Parameter] | None = None,
        asyncapi_extra: dict[str, Any] | None = None,
        **options: Any,
    ) -> None:
        if event_type is None:
            msg = "Event type is required"
            raise ValueError(msg)
        topic = topic or event_type.get_default_topic()
        if not topic:
            msg = "Topic is required"
            raise ValueError(msg)
        if concurrency < 1:
            msg = "Concurrency must be greater than 0"
            raise ValueError(msg)
        self.name = name
        self.event_type = event_type
        self.topic = topic
        self.timeout = timeout
        self.tags = tags
        self.encoder = encoder
        self.decoder = decoder
        self.dynamic = dynamic
        self.concurrency = concurrency
        self.retry_strategy = retry_strategy
        self.store_results = store_results
        self.parameters = parameters or {}
        self.description = description
        self.publishes = publishes or []
        self.asyncapi_extra = asyncapi_extra or {}
        self.options = options
        self.logger = get_logger(__name__, self.name)

    def maybe_set_publisher(self, publisher: Publisher) -> None:
        pass

    if TYPE_CHECKING:
        process: Callable[Concatenate[CloudEventType, ...], Awaitable[Any]]
    else:

        @abstractmethod
        async def process(self, message: CloudEventType) -> Any:
            raise NotImplementedError


class FnConsumer(Consumer[CloudEventType], Generic[CloudEventType, P]):
    def __init__(
        self,
        *,
        fn: Callable[Concatenate[CloudEventType, P], Awaitable[Any]],
        **extra: Any,
    ) -> None:
        if "name" not in extra:
            extra["name"] = fn.__name__
        if "event_type" not in extra:
            extra["event_type"] = resolve_message_type_hint(fn)
        if "description" not in extra:
            extra["description"] = fn.__doc__ or ""
        if not is_async_callable(fn):
            fn = to_async(fn)
        self.fn = resolved_func(fn)
        super().__init__(**extra)

    async def process(
        self, message: CloudEventType, *args: P.args, **kwargs: P.kwargs
    ) -> Any:
        return await self.fn(message, *args, **kwargs)


class GenericConsumer(Consumer[CloudEventType], ABC):
    def __init__(self, **extra: Any) -> None:
        if "name" not in extra:
            extra["name"] = getattr(type(self), "name", type(self).__name__)
        if "event_type" not in extra:
            extra["event_type"] = type(self).__orig_bases__[0].__args__[0]  # type: ignore[attr-defined]
        if "description" not in extra:
            extra["description"] = type(self).__doc__ or ""
        super().__init__(**extra)
        self._publish: Publisher | None = None
        self.process = resolved_func(self.process)

    def maybe_set_publisher(self, publisher: Publisher) -> None:
        self._publish = publisher

    @property
    def publish(self) -> Publisher:
        if self._publish is None:
            err = "Publisher is not set"
            raise RuntimeError(err)
        return self._publish


class ChannelConsumer(Consumer[CloudEventType]):
    def __inii__(
        self,
        channel: MemoryObjectSendStream[tuple[CloudEventType, Callable[[], None]]],
        **extra: Any,
    ) -> None:
        if "name" not in extra:
            extra["name"] = f"{socket.gethostname()}:{uuid4()}"
        super().__init__(**extra)
        self.channel = channel
        self._timeout = to_float(self.timeout) or 10.0

    async def process(self, message: CloudEventType) -> Any:
        event = anyio.Event()
        await self.channel.send((message, event.set))
        with anyio.fail_after(self._timeout):
            await event.wait()


class ConsumerGroup:
    def __init__(self, **options: Unpack[ConsumerGroupOptions]) -> None:
        self.options = options
        self.consumers: dict[str, Consumer] = {}

    def add_consumer(self, consumer: Consumer) -> None:
        self.consumers[consumer.name] = consumer

    def add_consumer_group(self, other: ConsumerGroup) -> None:
        self.consumers.update(other.consumers)

    @overload
    def subscribe(self, func_or_cls: MessageHandler) -> MessageHandler: ...

    @overload
    def subscribe(
        self,
        func_or_cls: None = None,
        name: str | None = None,
        event_type: type[CloudEventType] | None = None,
        topic: str | None = None,
        concurrency: int | None = None,
        timeout: Timeout | None = None,
        description: str | None = None,
        encoder: Encoder | None = None,
        decoder: Decoder | None = None,
        dynamic: bool | None = None,
        tags: list[str] | None = None,
        publishes: list[Publishes] | None = None,
        parameters: dict[str, Parameter] | None = None,
        asyncapi_extra: dict[str, Any] | None = None,
        **options: Any,
    ) -> Callable[[MessageHandler], MessageHandler]: ...

    def subscribe(
        self,
        func_or_cls: MessageHandler | None = None,
        name: str | None = None,
        event_type: type[CloudEventType] | None = None,
        topic: str | None = None,
        concurrency: int | None = None,
        timeout: Timeout | None = None,
        description: str | None = None,
        encoder: Encoder | None = None,
        decoder: Decoder | None = None,
        dynamic: bool | None = None,
        tags: list[str] | None = None,
        publishes: list[Publishes] | None = None,
        parameters: dict[str, Parameter] | None = None,
        asyncapi_extra: dict[str, Any] | None = None,
        **options: Any,
    ) -> MessageHandler | Callable[[MessageHandler], MessageHandler]:
        def decorator(func_or_cls: MessageHandler) -> MessageHandler:
            cls: type[Consumer] = FnConsumer
            if inspect.isfunction(func_or_cls):
                options["fn"] = func_or_cls

            elif isinstance(func_or_cls, type) and issubclass(
                func_or_cls,
                GenericConsumer,
            ):
                cls = func_or_cls
            else:
                msg = f"Expected function or GenericConsumer got {type(func_or_cls)}"
                raise TypeError(
                    msg,
                )
            options.update(
                {
                    "name": name,
                    "event_type": event_type,
                    "topic": topic,
                    "concurrency": concurrency,
                    "timeout": timeout,
                    "description": description,
                    "encoder": encoder,
                    "dynamic": dynamic,
                    "tags": tags,
                    "publishes": publishes,
                    "parameters": parameters,
                    "asyncapi_extra": asyncapi_extra,
                }
            )
            filtered_options = {k: v for k, v in options.items() if v is not None}
            for k, v in self.options.items():
                filtered_options.setdefault(k, v)
            consumer = cls(**filtered_options)
            self.add_consumer(consumer)
            return func_or_cls

        if func_or_cls is None:
            return decorator

        return decorator(func_or_cls)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic

from .logging import LoggerMixin
from .types import CloudEventType

if TYPE_CHECKING:
    from .consumer import Consumer
    from .exceptions import Fail, Retry, Skip
    from .service import Service


class Middleware(Generic[CloudEventType], LoggerMixin):
    """Base class for middlewares."""

    requires: type[CloudEventType] | None = None

    def __init__(self, service: Service) -> None:
        self.service = service

    async def before_broker_connect(self) -> None:
        """Called before broker connects."""

    async def after_broker_connect(self) -> None:
        """Called after broker connects."""

    async def before_broker_disconnect(self) -> None:
        """Called before broker disconnects."""

    async def after_broker_disconnect(self) -> None:
        """Called after broker disconnects."""

    async def before_consumer_start(
        self,
        *,
        consumer: Consumer,
    ) -> None:
        """Called before consumer is started."""

    async def after_consumer_start(
        self,
        *,
        consumer: Consumer,
    ) -> None:
        """Called after consumer is started."""

    async def before_ack(
        self,
        *,
        consumer: Consumer,
        raw_message: Any,
    ) -> None:
        """Called before message is acknowledged."""

    async def after_ack(
        self,
        *,
        consumer: Consumer,
        raw_message: Any,
    ) -> None:
        """Called after message is acknowledged."""

    async def before_nack(
        self,
        *,
        consumer: Consumer,
        raw_message: Any,
    ) -> None:
        """Called before message is rejected."""

    async def after_nack(
        self,
        *,
        consumer: Consumer,
        raw_message: Any,
    ) -> None:
        """Called after message is rejected."""

    async def before_publish(
        self,
        *,
        message: CloudEventType,
        **kwargs: Any,
    ) -> None:
        """Called before message is published."""

    async def after_publish(
        self,
        *,
        message: CloudEventType,
        **kwargs: Any,
    ) -> None:
        """Called after message is published."""

    async def after_skip_message(
        self,
        *,
        consumer: Consumer,
        message: CloudEventType,
        exc: Skip,
    ) -> None:
        """Called after message is skipped by the middleware."""

    async def after_fail_message(
        self,
        *,
        consumer: Consumer,
        message: CloudEventType,
        exc: Fail,
    ) -> None:
        """Called after message is failed by the middleware."""

    async def after_retry_message(
        self,
        *,
        consumer: Consumer,
        message: CloudEventType,
        exc: Retry,
    ) -> None:
        """Called after message is retried by the middleware."""

    async def before_process_message(
        self,
        *,
        consumer: Consumer,
        message: CloudEventType,
    ) -> None:
        """Called before message is processed."""

    async def after_process_message(
        self,
        *,
        consumer: Consumer,
        message: CloudEventType,
        result: Any = None,
        exc: Exception | None = None,
    ) -> None:
        """Called after message is processed (but not acknowledged/rejected yet)."""



from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Any, Union

import aio_pika
from aio_pika.abc import (
    AbstractExchange,
    AbstractIncomingMessage,
    AbstractRobustConnection,
    DateType,
)
from aiormq.abc import ConfirmationFrameType
from anyio import move_on_after
from pydantic import AnyUrl, UrlConstraints

from eventiq.broker import UrlBroker
from eventiq.settings import UrlBrokerSettings

if TYPE_CHECKING:
    from datetime import datetime

    from anyio.streams.memory import MemoryObjectSendStream

    from eventiq import Consumer
    from eventiq.types import ID, DecodedMessage


RabbitmqUrl = Annotated[AnyUrl, UrlConstraints(allowed_schemes=["amqp"])]


class RabbitMQSettings(UrlBrokerSettings[RabbitmqUrl]):
    default_prefetch_count: int = 10
    exchange_name: str = "default"


class RabbitmqBroker(
    UrlBroker[AbstractIncomingMessage, Union[ConfirmationFrameType, None]]
):
    """
    RabbitMQ broker implementation, based on `aio_pika` library.
    :param default_prefetch_count: default number of messages to prefetch (per queue)
    :param queue_options: additional queue options
    :param exchange_name: global exchange name
    :param kwargs: Broker base class parameters
    """

    Settings = RabbitMQSettings
    protocol = "amqp"
    WILDCARD_ONE = "*"
    WILDCARD_MANY = "#"

    def __init__(
        self,
        *,
        default_prefetch_count: int = 10,
        queue_options: dict[str, Any] | None = None,
        exchange_name: str = "events",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.default_prefetch_count = default_prefetch_count
        self.queue_options = queue_options or {}
        self.exchange_name = exchange_name
        self._connection: AbstractRobustConnection | None = None
        self._exchange: AbstractExchange | None = None

    @property
    def connection(self) -> AbstractRobustConnection:
        if self._connection is None:
            raise self.connection_error
        return self._connection

    @property
    def exchange(self) -> AbstractExchange:
        if self._exchange is None:
            raise self.connection_error
        return self._exchange

    def should_nack(self, raw_message: AbstractIncomingMessage) -> bool:
        return bool(raw_message.redelivered)

    async def connect(self) -> None:
        self._connection = await aio_pika.connect_robust(
            self.url, **self.connection_options
        )
        channel = await self.connection.channel()
        self._exchange = await channel.declare_exchange(
            name=self.exchange_name, type=aio_pika.ExchangeType.TOPIC, durable=True
        )

    async def disconnect(self) -> None:
        await self.connection.close()

    @property
    def is_connected(self) -> bool:
        return not self.connection.is_closed

    @staticmethod
    def decode_message(raw_message: AbstractIncomingMessage) -> DecodedMessage:
        return raw_message.body, {k: str(v) for k, v in raw_message.headers.items()}

    @staticmethod
    def get_message_metadata(raw_message: AbstractIncomingMessage) -> dict[str, str]:
        return {}

    async def sender(
        self, group: str, consumer: Consumer, send_stream: MemoryObjectSendStream
    ) -> None:
        channel = await self.connection.channel()
        prefetch_count = consumer.concurrency * 2
        await channel.set_qos(prefetch_count=prefetch_count)
        options: dict[str, Any] = consumer.options.get(
            "queue_options", self.queue_options
        )
        is_durable = not consumer.dynamic
        options.setdefault("durable", is_durable)
        queue = await channel.declare_queue(name=f"{group}:{consumer.name}", **options)
        await queue.bind(self.exchange, routing_key=consumer.topic)
        try:
            async with send_stream, queue.iterator() as q:
                async for message in q:
                    await send_stream.send(message)

        finally:
            with move_on_after(1, shield=True):
                if consumer.dynamic:
                    await queue.unbind(self.exchange, routing_key=consumer.topic)
                await channel.close()

    async def publish(
        self,
        topic: str,
        body: bytes,
        *,
        headers: dict[str, str],
        message_id: ID,
        message_type: str,
        message_content_type: str,
        message_time: datetime,
        message_source: str,
        timeout: float | None = None,
        mandatory: bool = True,
        immidiate: bool = False,
        delivery_mode: aio_pika.DeliveryMode = aio_pika.DeliveryMode.PERSISTENT,
        priority: int | None = None,
        correlation_id: str | None = None,
        reply_to: str | None = None,
        expiration: DateType | None = None,
        user_id: str | None = None,
        **kwargs: Any,
    ) -> ConfirmationFrameType | None:
        msg = aio_pika.Message(
            body,
            headers=dict(headers),
            content_type=message_content_type,
            content_encoding="UTF-8",
            delivery_mode=delivery_mode,
            priority=priority,
            correlation_id=correlation_id,
            reply_to=reply_to,
            expiration=expiration,
            message_id=str(message_id),
            timestamp=message_time,
            type=message_type,
            user_id=user_id,
            app_id=message_source,
        )
        return await self.exchange.publish(
            msg,
            routing_key=topic,
            mandatory=mandatory,
            immediate=immidiate,
            timeout=timeout,
        )

    async def ack(self, raw_message: AbstractIncomingMessage) -> None:
        await raw_message.ack()

    async def nack(
        self, raw_message: AbstractIncomingMessage, delay: int | None = None
    ) -> None:
        await raw_message.reject(requeue=True)



from __future__ import annotations

import asyncio
from pathlib import Path
from typing import TYPE_CHECKING

from eventiq.middleware import Middleware

if TYPE_CHECKING:
    from eventiq import Service


class HealthCheckMiddleware(Middleware):
    """Middleware for performing basic health checks on broker."""

    def __init__(
        self,
        service: Service,
        interval: int = 5,
        base_dir: str = "/tmp",  # nosec
    ) -> None:
        super().__init__(service)
        self.base_dir = Path(base_dir)
        self.interval = interval
        self._task: asyncio.Task | None = None

    async def after_broker_connect(self) -> None:
        self._task = asyncio.create_task(self._run_forever())

    async def _run_forever(self) -> None:
        p = self.base_dir / "healthy"
        broker = self.service.broker
        while True:
            try:
                unhealthy = not broker.is_connected
            except Exception as e:
                self.logger.exception("Healthcheck failed", exc_info=e)
                unhealthy = True

            if unhealthy and p.exists():
                p.rename(self.base_dir / "unhealthy")
            else:
                p.touch(exist_ok=True)
            await asyncio.sleep(self.interval)

    async def after_broker_disconnect(self) -> None:
        if self._task:
            self._task.cancel()
            self._task = None
